#!/usr/bin/python
#
# 15/04/2018
#
# backup-vms.py - this script is used to backup vms from nuxis3/oVirt
# 
# cmar@eurotux.com
#
#
# based on:
#     https://github.com/oVirt/ovirt-engine-sdk/blob/master/sdk/examples/vm_backup.py
#     https://github.com/oVirt/ovirt-engine-sdk/blob/master/sdk/examples/download_disk_snapshots.py
#     https://github.com/wefixit-AT/oVirtBackup

from ConfigParser import (
    RawConfigParser,
    NoSectionError,
    NoOptionError,
)

import os
import sys
import time
import uuid
import ssl
import logging
import json
import string
import random
import fnmatch

from time import strftime
from argparse import ArgumentParser, FileType

import ovirtsdk4 as sdk
import ovirtsdk4.types as types

from httplib import HTTPSConnection

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

# logger
logger = logging.getLogger()

# The name of the application, to be used as the 'origin' of events
# sent to the audit log:
APPLICATION_NAME = 'backup-vms'

CONFIG_SECTION = "config"
DEFAULTS = {
    "logger_fmt": "%(asctime)s: %(message)s",
    "logger_file_path": None,
}

class Config(object):
    """
    Class to read the config from the config file and serve these config
    """
    def __init__(self, fd, debug, arguments):
        try:
            self._cp = config_parser = RawConfigParser(defaults=DEFAULTS)
            config_parser.readfp(fd)

            section = CONFIG_SECTION
            # Update with options passed from CLI interface
            for key, val in arguments.items():
                if val is not None:
                    if isinstance(val, basestring):
                        config_parser.set(section, key, str(val))
                    else:
                        config_parser.set(section, key, json.dumps(val))

            self.__ca = config_parser.get(section, "ca")

            self.__server = config_parser.get(section, "server")
            self.__username = config_parser.get(section, "username")
            self.__password = config_parser.get(section, "password")

            self.__vm_names = json.loads(config_parser.get(section, "vm_names"))
            self.__debug = debug
            self.__logger_fmt = config_parser.get(section, "logger_fmt")
            self.__logger_file_path = config_parser.get(section, "logger_file_path")

            self.__backups_dir = config_parser.get(section, "backups_dir")

            try:
                self.__keep_days = config_parser.get(section, "keep_days")
            except:
                self.__keep_days = None

        except (NoSectionError, NoOptionError) as e:
            print str(e)
            sys.exit(1)

    def get_vm_names(self):

        return self.__vm_names

    def get_ca(self):
        return self.__ca

    def get_server(self):
        return self.__server

    def get_username(self):
        return self.__username

    def get_password(self):
        return self.__password

    def get_debug(self):
        return self.__debug

    def get_logger_fmt(self):
        return self.__logger_fmt

    def get_logger_file_path(self):
        return self.__logger_file_path

    def get_backups_dir(self):
        return self.__backups_dir

    def get_keep_days(self):
        return int(self.__keep_days)

# Create the connection to the server:
def get_connection(url, username, password, ca, debug=False):
    return sdk.Connection(
        url=url,
        username=username,
        password=password,
        ca_file=ca,
        debug=debug,
        log=logging.getLogger(),
    )

# initialize logger
def initialize_logger(logger_fmt, logger_file_path, debug):
    logger_options = {
        "format": logger_fmt,
        "level": logging.DEBUG if debug else logging.INFO,
    }
    if logger_file_path:
        logger_options['filename'] = logger_file_path
    logging.basicConfig(**logger_options)

def create_argparser():
    p = ArgumentParser()
    # General options
    p.add_argument(
        "-c", "--config-file",
        help="Path to the config file, pass dash (-) for stdin",
        dest="config_file",
        #required=True,
        type=FileType(),
    )
    p.add_argument(
        "--backups-dir",
        help="Path to the backups directory",
        dest="backups_dir",
        #required=True,
        #type=DirType(),
    )
    p.add_argument(
        "-d", "--debug",
        help="Debug flag",
        dest="debug",
        action="store_true",
        default=False,
    )
    p.add_argument(
        "--keep-days",
        help="Number of days to keeo backups",
        dest="keep_days",
        default=None,
    )

    osg = p.add_argument_group("oVirt server related options")
    osg.add_argument(
        "--server",
        help="URL to connect to your engine",
        dest="server",
        default=None,
    )
    osg.add_argument(
        "--username",
        help="Username to connect to the engine",
        dest="username",
        default=None,
    )
    osg.add_argument(
        "--password",
        help="Password to connect to the engine",
        dest="password",
        default=None,
    )
    osg.add_argument(
        "--ca",
        help="CA certificate file path",
        dest="ca",
        default=None,
    )

    vmg = p.add_argument_group("VM's related arguments")
    vmg.add_argument(
        "-a", "--all-vms",
        help="Backup all VMs and override the list of VM's in the config "
        "file",
        dest="all_vms",
        action="store_true",
        default=False,
    )
    vmg.add_argument(
        "--vm-names",
        help="List of names which VMs should be backed up",
        dest="vm_names",
        default=None,
        nargs='+'
    )

    lg = p.add_argument_group("Logging related options")
    lg.add_argument(
        "--logger-fmt",
        help="This value is used to format log messages",
        dest="logger_fmt",
        default=None,
    )
    lg.add_argument(
        "--logger-file-path",
        help="Path to file where we to store log messages",
        dest="logger_file_path",
        default=None,
    )
    return p

def get_transfer_service(disk_snapshot_id):
    # Get a reference to the service that manages the image transfer:
    transfers_service = system_service.image_transfers_service()

    # Add a new image transfer:
    transfer = transfers_service.add(
        types.ImageTransfer(
            snapshot=types.DiskSnapshot(id=disk_snapshot_id),
            direction=types.ImageTransferDirection.DOWNLOAD,
        )
    )

    # Get reference to the created transfer service:
    transfer_service = transfers_service.image_transfer_service(transfer.id)

    while transfer.phase == types.ImageTransferPhase.INITIALIZING:
        time.sleep(1)
        transfer = transfer_service.get()

    return transfer_service

def get_proxy_connection(proxy_url):
    # At this stage, the SDK granted the permission to start transferring the disk, and the
    # user should choose its preferred tool for doing it - regardless of the SDK.
    # In this example, we will use Python's httplib.HTTPSConnection for transferring the data.
    context = ssl.create_default_context()

    # Note that ovirt-imageio-proxy by default checks the certificates, so if you don't have
    # your CA certificate of the engine in the system, you need to pass it to HTTPSConnection.
    context.load_verify_locations(cafile=config.get_ca())

    return HTTPSConnection(
        proxy_url.hostname,
        proxy_url.port,
        context=context,
    )

def download_disk_snapshot(snapshot,disk,vm_backup_dir):

    logger.info("Downloading snapshot '%s' of disk '%s'('%s')" % (snapshot.id,disk.alias,disk.id))

    transfer_service = None
    try:
        transfer_service = get_transfer_service(snapshot.id)
        transfer = transfer_service.get()
        proxy_url = urlparse(transfer.proxy_url)
        proxy_connection = get_proxy_connection(proxy_url)

        # put disk in subdir
        disk_dir = "%s/%s-%s" % (vm_backup_dir,disk.alias,disk.id)
        if not os.path.exists(disk_dir):
            try:
                os.makedirs(disk_dir)
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise

        path = "%s/%s" % (disk_dir,snapshot.id)

        with open(path, "wb") as mydisk:
            # Set needed headers for downloading:
            transfer_headers = {
                'Authorization': transfer.signed_ticket,
            }

            # Perform the request.
            proxy_connection.request(
                'GET',
                proxy_url.path,
                headers=transfer_headers,
            )
            # Get response
            r = proxy_connection.getresponse()

            # Check the response status:
            if r.status >= 300:
                logger.error("Error: %s" % r.read())

            bytes_to_read = int(r.getheader('Content-Length'))
            #chunk_size = 64 * 1024 * 1024
            chunk_size = 64 * 1024

            logger.info("   Disk snapshot size: %s bytes" % str(bytes_to_read))

            while bytes_to_read > 0:
                # Calculate next chunk to read
                to_read = min(bytes_to_read, chunk_size)

                # Read next chunk
                chunk = r.read(to_read)

                if chunk == "":
                    raise RuntimeError("Socket disconnected")

                # Write the content to file:
                mydisk.write(chunk)

                # Update bytes_to_read
                bytes_to_read -= len(chunk)

                completed = 1 - (bytes_to_read / float(r.getheader('Content-Length')))

                logger.debug("Completed", "{:.0%}".format(completed))
    finally:
        # Finalize the session.
        if transfer_service is not None:
            transfer_service.finalize()

def delete_old_backups(base_dir,vm,days):
    timedays = time.time() - days*24*60*60
    pattern = "*/%s-%s" % (vm.name,vm.id)
    for root, dirs, files in os.walk(base_dir):
        if fnmatch.fnmatch(root, pattern):
            for name in files:
                fpath = "%s/%s" % (root,name)
                if os.path.getmtime(fpath) < timedays:
                    logger.info("Delete old backup %s" % fpath)
                    os.remove(fpath)

            if not os.listdir(root):
                os.rmdir(root)
                

def arguments_to_dict(opts):
    result = {}
    ignored_keys = ('config_file', 'debug')
    for key, val in vars(opts).items():
        if key in ignored_keys:
            continue  # These doesn't have a place in config file
        if val is not None:
            result[key] = val
    return result

# generate random token
def rand_tok(size=6, chars=string.hexdigits):
    return ''.join(random.choice(chars) for _ in range(size))

# main function
def main(argv):

    # parse args
    p = create_argparser()
    # options
    opts = p.parse_args(argv)

    # load config file
    config_arguments = arguments_to_dict(opts)

    global config
    with opts.config_file:
        config = Config(opts.config_file, opts.debug, config_arguments)

    # init logger
    initialize_logger(
        config.get_logger_fmt(), config.get_logger_file_path(), opts.debug,
    )

    # Create a connection to the server:
    connection = get_connection( url=config.get_server(),
                                 username=config.get_username(),
                                 password=config.get_password(),
                                 ca=config.get_ca(),
                                 debug=config.get_debug())

    global system_service
    # Get a reference to the root service:
    system_service = connection.system_service()

    # Get the reference to the "vms" service:
    vms_service = system_service.vms_service()

    # Get a reference to the storage domains service:
    storage_domains_service = system_service.storage_domains_service()

    # Get the reference to the service that we will use to send events to
    # the audit log:
    events_service = system_service.events_service()

    # In order to send events we need to also send unique integer ids.
    event_id = int(time.time())

    base_backups_dir = "%s/%s%s" % (config.get_backups_dir(),strftime("%Y%m%d"),rand_tok())
    if not os.path.exists(base_backups_dir):
        try:
            os.makedirs(base_backups_dir)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise

    # select VMs to backup
    try:
        backup_list = []
        if opts.all_vms:
            # list all VMs
            backup_list = vms_service.list()
        else:
            for vm_name in config.get_vm_names():
                # filter VMs by name
                res = vms_service.list(search="name=%s" % str(vm_name), all_content=True)
                if len(res) > 0:
                    backup_list.append(res[0])
                else:
                    logger.error("Unable to find VM '%s'" % str(vm_name))
        if len(backup_list) > 0:
            for vm in backup_list:

                # delete old backups
                if config.get_keep_days():
                    delete_old_backups(config.get_backups_dir(),vm,config.get_keep_days())

                vm_backup_dir = "%s/%s-%s" % (base_backups_dir,vm.name,vm.id)
                if not os.path.exists(vm_backup_dir):
                    try:
                        os.makedirs(vm_backup_dir)
                    except OSError as e:
                        if e.errno != errno.EEXIST:
                            raise

                logger.info("Backup VM '%s' with id '%s' to '%s'" % (vm.name, vm.id, vm_backup_dir))

                # Create an unique description for the snapshot, so that it is easier
                # for the administrator to identify this snapshot as a temporary one
                # created just for backup purposes:
                snap_description = '%s-backup-%s' % (vm.name, uuid.uuid4())
                
                # Send an external event to indicate to the administrator that the
                # backup of the virtual machine is starting. Note that the description
                # of the event contains the name of the virtual machine and the name of
                # the temporary snapshot, this way, if something fails, the administrator
                # will know what snapshot was used and remove it manually.
                events_service.add(
                    event=types.Event(
                        vm=types.Vm(
                          id=vm.id,
                        ),
                        origin=APPLICATION_NAME,
                        severity=types.LogSeverity.NORMAL,
                        custom_id=event_id,
                        description=(
                            'Backup of virtual machine \'%s\' using snapshot \'%s\' is '
                            'starting.' % (vm.name, snap_description)
                        ),
                    ),
                )
                event_id += 1

                # Save the OVF to a file, so that we can use to restore the virtual
                # machine later. The name of the file is the name of the virtual
                # machine, followed by a dash and the identifier of the virtual machine,
                # to make it unique:
                ovf_data = vm.initialization.configuration.data
                ovf_file = '%s/%s-%s.ovf' % (vm_backup_dir,vm.name, vm.id)
                with open(ovf_file, 'w') as ovs_fd:
                    ovs_fd.write(ovf_data.encode('utf-8'))
                logging.info('Wrote OVF to file \'%s\'.', os.path.abspath(ovf_file))

                # Locate the service that manages the virtual machine:
                vm_service = vms_service.vm_service(vm.id)
                
                # Send the request to create the snapshot. Note that this will return
                # before the snapshot is completely created, so we will later need to
                # wait till the snapshot is completely created.
                # The snapshot will not include memory. Change to True the parameter
                # persist_memorystate to get it (in that case the VM will be paused for a while).
                snaps_service = vm_service.snapshots_service()

                if len(snaps_service.list()) > 1:
                    logger.error("VM '%s' already have snapshosts and this is not supported" % (vm.name))
                else:
                    snap = snaps_service.add(
                        snapshot=types.Snapshot(
                            description=snap_description,
                            persist_memorystate=False,
                        ),
                    )
                    logging.info(
                        'Sent request to create snapshot \'%s\', the id is \'%s\'.',
                        snap.description, snap.id,
                    )

                    # Poll and wait till the status of the snapshot is 'ok', which means
                    # that it is completely created:
                    snap_service = snaps_service.snapshot_service(snap.id)
                    while snap.snapshot_status != types.SnapshotStatus.OK:
                        logging.debug(
                            'Waiting till the snapshot is created, the status is now \'%s\'.',
                            snap.snapshot_status )
                        time.sleep(1)
                        snap = snap_service.get()
                        time.sleep(1)

                    # add some delay
                    time.sleep(10)

                    logging.info('  The snapshot is now complete.')

                    # Retrieve the descriptions of the disks of the snapshot:
                    snap_disks_service = snap_service.disks_service()
                    snap_disks = snap_disks_service.list()

                    # download disk snaphost
                    for snap_disk in snap_disks:
                        disk_id = snap_disk.id
                        sd_id = snap_disk.storage_domains[0].id

                        # Get a reference to the storage domain service in which the disk snapshots reside:
                        storage_domain_service = storage_domains_service.storage_domain_service(sd_id)

                        # Get a reference to the disk snapshots service:
                        disk_snapshot_service = storage_domain_service.disk_snapshots_service()

                        # Get a list of disk snapshots by a disk ID
                        all_disk_snapshots = disk_snapshot_service.list()
                    
                        # Filter disk snapshots list by disk id
                        disk_snapshots = [s for s in all_disk_snapshots if s.disk.id == disk_id]
                    
                        # Download disk snapshot
                        if len(disk_snapshots) > 0:
                            for disk_snapshot in disk_snapshots:
                                download_disk_snapshot(disk_snapshot,snap_disk,vm_backup_dir)

                    # Remove the snapshot:
                    snap_service.remove()
                    logging.info('Removed the snapshot \'%s\'.', snap.description)
                    
                    # Send an external event to indicate to the administrator that the
                    # backup of the virtual machine is completed:
                    events_service.add(
                        event=types.Event(
                            vm=types.Vm(
                              id=vm.id,
                            ),
                            origin=APPLICATION_NAME,
                            severity=types.LogSeverity.NORMAL,
                            custom_id=event_id,
                            description=(
                                'Backup of virtual machine \'%s\' using snapshot \'%s\' is '
                                'completed.' % (vm.name, snap_description)
                            ),
                        ),
                    )
                    event_id += 1
        else:
            logger.error("No VMs to backup.")
    finally:
        # Close the connection to the server:
        connection.close()

if __name__ == "__main__":
    main(sys.argv[1:])
